import Flutter
import UIKit
import AliyunOSSiOS
import MobileCoreServices

public class SwiftAliyunOSSPlugin: NSObject, FlutterPlugin, UIDocumentPickerDelegate, UINavigationControllerDelegate {
  
  let channel: FlutterMethodChannel
  var flutterViewController: UIViewController?
  var client: OSSClient!
  var bucketName: String!
  var uploadTaskDatas: [UploadTaskData] = []
  
  init(channel: FlutterMethodChannel) {
    self.channel = channel
  }
  
  public static func register(with registrar: FlutterPluginRegistrar) {
    let channel = FlutterMethodChannel(name: "com.wombat/aliyun_oss_plugin", binaryMessenger: registrar.messenger())
    let instance = SwiftAliyunOSSPlugin(channel: channel)
    registrar.addMethodCallDelegate(instance, channel: channel)
  }
  
  public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
    switch call.method{
    case "init":
      initClient(call)
      result(["init_result": true])
    case "upload":
      self.bucketName = (call.arguments as! [String: String])["bucketName"]
      selectFile()
      
    case "uploadWithPath":
      let params = call.arguments as! [String: String]
      self.bucketName = params["bucketName"]
      let filePath: String = params["filePath"]!
      let md5: String = params["md5"]!
      let fileURL = URL(fileURLWithPath: filePath)
      resumableUpload(fileURL: fileURL, objectKey: "\(md5).\(fileURL.pathExtension)")
                  
    case "getMd5Str":
      let filePath = (call.arguments as! [String: String])["filePath"]
      DispatchQueue(label: "calcuate Md5", qos: .background).async {
        print("正在计算md5中...")
        let start = CACurrentMediaTime()
        guard let md5: String = OSSUtil.fileMD5String(filePath!) else {
          print("md5内容为空")
          result("")
          return
        }
        let elapsed = CACurrentMediaTime() - start
        print("计算md5: \(md5), 耗时: \(elapsed)")
        DispatchQueue.main.async {
          result(md5)
        }
      }
    case "pause_upload":
      let objectKey = (call.arguments as! [String: String])["objectKey"]
      let pauseResult = pauseUpload(objectKey!)
      result(pauseResult)
    case "resume_upload":
      let objectKey = (call.arguments as! [String: String])["objectKey"]
      let resumeResult = resumeUpload(objectKey!)
      result(resumeResult)
    case "cancel_upload":
      let objectKey = (call.arguments as! [String: String])["objectKey"]
      let resumeResult = cancelUpload(objectKey!)
      result(objectKey)
    case "cancel_all":
      cancelAllUpload()
      result(bucketName)
    default:
      result(FlutterMethodNotImplemented)
    }
  }
  
  func initClient(_ call: FlutterMethodCall) {
    let params = call.arguments as! [String: Any?]
    let endpoint = params["endpoint"] as! String
    let accessKeyId = params["accessKeyId"] as! String
    let secretKeyId = params["secretKeyId"] as! String
    let securityToken = params["securityToken"] as? String
    
    var credential: OSSCredentialProvider
    if (securityToken == nil) {
      credential = OSSPlainTextAKSKPairCredentialProvider(plainTextAccessKey: accessKeyId, secretKey: secretKeyId)
    } else {
      credential = OSSStsTokenCredentialProvider(accessKeyId: accessKeyId, secretKeyId: secretKeyId, securityToken: securityToken!)
    }
    
    client = OSSClient(endpoint: endpoint, credentialProvider: credential)
    
  }
  
  func selectFile() {
    let documentPicker: UIDocumentPickerViewController = UIDocumentPickerViewController(documentTypes: ["public.item"], in: UIDocumentPickerMode.import)
    documentPicker.delegate = self
    documentPicker.modalPresentationStyle = UIModalPresentationStyle.fullScreen
    if (flutterViewController == nil) {
      flutterViewController = UIApplication.shared.keyWindow?.rootViewController;
    }
    flutterViewController?.present(documentPicker, animated: true, completion: nil);
    
  }
  
  public func documentPicker(_ controller: UIDocumentPickerViewController, didPickDocumentsAt urls: [URL]) {
    guard let fileURL = urls.first else {
      return
    }
    self.uploadFile(fileURL: fileURL)
  }
  
  public func documentPickerWasCancelled(_ controller: UIDocumentPickerViewController) {
    flutterViewController?.dismiss(animated: true, completion: nil)
  }
  
  func uploadFile(fileURL: URL) {
    DispatchQueue(label: "calcuate Md5", qos: .background).async {
      print("正在计算md5中...")
      let start = CACurrentMediaTime()
      guard let md5: String = OSSUtil.fileMD5String(fileURL.path) else {
        print("md5内容为空")
        return
      }
      let elapsed = CACurrentMediaTime() - start
      print("计算md5: \(md5), 线程: \(Thread.current), 耗时: \(elapsed)")
      let ext = ".\(fileURL.pathExtension)"
      let objectKey = "\(md5)\(ext)"
      print("文件路径: \(fileURL.path), \(objectKey)")
      DispatchQueue.main.async {
        self.resumableUpload(fileURL: fileURL, objectKey: objectKey)
      }
    }
  }
  
  func pauseUpload(_ objectKey: String)-> String? {
    if let taskData = uploadTaskDatas.first(where: {
      $0.objectKey == objectKey
    }) {
      taskData.request.cancel()
      return objectKey
    }
    return nil
  }
  
  func resumeUpload(_ objectKey: String)-> String? {
    if let taskData = uploadTaskDatas.first(where: {
      $0.objectKey == objectKey
    }) {
      uploadTaskDatas.removeAll(where: {
        $0.objectKey == objectKey
      })
      resumableUpload(fileURL: taskData.getFileURL(), objectKey: taskData.objectKey)
      return objectKey
    }
    return nil
  }
  
  func cancelUpload(_ objectKey: String)-> String? {
    if let taskData = uploadTaskDatas.first(where: {
      $0.objectKey == objectKey
    }) {
      uploadTaskDatas.removeAll(where: {
        $0.objectKey == objectKey
      })
      
      taskData.request.cancel()
      return objectKey
    }
    return nil
  }
  
  func cancelAllUpload()-> String? {
    for taskData in uploadTaskDatas {
      taskData.request.cancel()
    }
    uploadTaskDatas.removeAll()
    return bucketName
  }
  
  func showUpload(task: OSSTask<AnyObject>?, objectKey: String, uploadId: String) -> Void {
    
    if (task?.error != nil) {
      let error: NSError = (task?.error)! as NSError
      print("上传出错: \(error)")
      uploadFailCallback(objectKey: objectKey, uploadId: uploadId, error: error)
    } else {
      let result = task?.result
      print("上传成功: \(result)")
      uploadSuccessCallback(objectKey: objectKey, uploadId: uploadId)
    }
  }
  
  func uploadInitCallback(filePath: String, objectKey: String) {
    channel.invokeMethod("upload_init", arguments: [
      "filePath": filePath,
      "objectKey": objectKey
    ])
  }
  
  func uploadProgressCallback(objectKey: String,
                              uploadId: String? = nil,
                              bytesSent: Int64,
                              totalBytesSent: Int64,
                              totalBytesExpectedToSend: Int64) {
    channel.invokeMethod("upload_progress", arguments: [
      "objectKey": objectKey,
      "uploadId": uploadId,
      "bytesSent": bytesSent,
      "totalBytesSent": totalBytesSent,
      "totalBytesExpectedToSend": totalBytesExpectedToSend,
    ]);
  }
  
  func uploadSuccessCallback(objectKey: String, uploadId: String?,result: AnyObject? = nil) {
    channel.invokeMethod("upload_success", arguments: [
      "objectKey": objectKey,
      "uploadId": uploadId,
      "result": result
    ])
  }
  
  func uploadFailCallback(objectKey: String, uploadId: String?, error: NSError) {
    channel.invokeMethod("upload_fail", arguments: [
      "objectKey": objectKey,
      "uploadId": uploadId,
      "error": error
    ])
  }
  
  func resumableUpload(fileURL: URL, objectKey: String) -> Void {
    let request = OSSResumableUploadRequest()
    /// 这个request无法添加objectMeta：用户自定义header数据
    request.uploadingFileURL = fileURL
    request.bucketName = bucketName
    request.deleteUploadIdOnCancelling = false;
    request.objectKey = objectKey
    let cacheDir =  NSSearchPathForDirectoriesInDomains(FileManager.SearchPathDirectory.cachesDirectory, FileManager.SearchPathDomainMask.userDomainMask, true).first
    request.recordDirectoryPath = cacheDir!
    request.partSize = 102400;
    request.uploadProgress = { (bytesSent: Int64, totalBytesSent: Int64, totalBytesExpectedToSend: Int64) -> Void in
      self.uploadProgressCallback(
        objectKey: request.objectKey,
        uploadId: request.uploadId,
        bytesSent: bytesSent,
        totalBytesSent: totalBytesSent,
        totalBytesExpectedToSend: totalBytesExpectedToSend)
    }
    
    let task = client.resumableUpload(request)
    
    uploadTaskDatas.append(UploadTaskData(
      objectKey: objectKey,
      uploadId: request.uploadId,
      bucketName: self.bucketName,
      filePath: fileURL.path,
      request: request))
    
    task.continue({ (t) -> Any? in
      self.showUpload(task: t, objectKey: request.objectKey, uploadId: request.uploadId)
    })
    
  }
  
}

